#include <../arch/arch.h>
#include <../arch/common/glctx.h>

vboolean autowarp = false;
#include "dumpevt.c"

#ifndef PFD_GENERIC_ACCELERATED
#define PFD_GENERIC_ACCELERATED         0x00001000
#endif	/* !PFD_GENERIC_ACCELERATED */

typedef struct win32_gl_context_s
{
	ARCH_COMMON_GL_CONTEXT_FIELDS
	HGLRC wgl_context;
}	win32_gl_context_t, *win32_gl_context_p;

void V_GLOBAL
win32_end_frame(win32_gl_context_p gl_context)
{
	SwapBuffers(gl_context->wnd->dc);
}

V_LOCAL void
win32_shutdown_gl(win32_gl_context_p gl_context)
{
	wglMakeCurrent(NULL, NULL);
	if(gl_context->wgl_context)
		wglDeleteContext(gl_context->wgl_context);
	ARCH_CLEAR_COMMON_GL_CONTEXT_FIELDS(gl_context);
	memFree(gl_context);
}

win32_gl_context_p
win32_init_gl(char *caption,
	int x, int y, int width, int height,
	vboolean fullscreen,
	vboolean can_resize,
	vboolean need_depth, vboolean need_stencil, vboolean need_dst_alpha)
{
	ARCH_CALLER_NAME(vglInit)
	PIXELFORMATDESCRIPTOR pfd;
	u32_t bpp;
	int i, pixfmt;
	GLfloat aspect;
	win32_gl_context_p gl_context;
	gl_context = (win32_gl_context_p)memAllocZ(sizeof(win32_gl_context_t));
	if(!gl_context)
		return false;
	gl_context->wnd = archWinOpenNP(NULL, NULL,
		caption, x, y, width, height, !fullscreen, can_resize);
	if(!gl_context->wnd)
		goto err_exit;
	archWinShow(gl_context->wnd, true);
	bpp = GetDeviceCaps(gl_context->wnd->dc, BITSPIXEL);
	if(bpp <= 8)
	{
		if(!fullscreen)
		{
			ARCH_ERROR("true or high color mode required");
			goto err_exit;
		}
		else
			bpp = 16;
	}
	if(fullscreen)
	{
		for(i = 0; i < gl_context->wnd->num_vidmodes; i++)
		{
			vidmode_p mode = gl_context->wnd->vidmodes + i;
			if((int)mode->width == width &&
					(int)mode->height == height && mode->bpp == bpp)
			{
				if(!archWinSetVidMode(gl_context->wnd, i))
				{
					ARCH_ERROR("archWinSetVidMode() failed");
					goto err_exit;
				}
				break;
			}
		}
		if(i == gl_context->wnd->num_vidmodes)
		{
			ARCH_ERROR("required video mode not found");
			goto err_exit;
		}
	}
#ifdef __BORLANDC__
	_control87(MCW_EM,MCW_EM);
#endif
	memZero(&pfd, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	pfd.nVersion = 1;
    pfd.iPixelType = PFD_TYPE_RGBA;
	pfd.iLayerType = PFD_MAIN_PLANE;
	pfd.cColorBits = 16;
    if(need_depth)
		pfd.cDepthBits = 16;
    if(need_stencil)
		pfd.cStencilBits = 1;
    if(need_dst_alpha)
		pfd.cAlphaBits = 1;
	pfd.dwFlags = PFD_DRAW_TO_WINDOW | PFD_DOUBLEBUFFER
		| PFD_SUPPORT_OPENGL;
	pfd.iPixelType = PFD_TYPE_RGBA;
	pixfmt = ChoosePixelFormat(gl_context->wnd->dc, &pfd);
	if(!pixfmt)
	{
		SYSCALL_FAILED("ChoosePixelFormat");
		win32_shutdown_gl(gl_context);
		return false;
	}
	memZero(&pfd, sizeof(PIXELFORMATDESCRIPTOR));
	pfd.nSize = sizeof(PIXELFORMATDESCRIPTOR);
	if(!DescribePixelFormat(gl_context->wnd->dc, pixfmt, pfd.nSize, &pfd))
	{
		SYSCALL_FAILED("DescribePixelFormat");
		goto err_exit;
	}
	if(!(pfd.dwFlags & PFD_GENERIC_ACCELERATED))
	{
		ARCH_WARNING("using software renderer");
	}
	if(need_depth && !pfd.cDepthBits)
	{
		ARCH_ERROR("renderer doesn't support Z-buffering");
		goto err_exit;
	}
	if(need_dst_alpha && !pfd.cAlphaBits)
	{
		ARCH_ERROR("renderer doesn't support dst alpha");
		goto err_exit;
	}
	if(need_stencil && !pfd.cStencilBits)
	{
		ARCH_ERROR("renderer doesn't support stencil");
		goto err_exit;
	}
	if(!SetPixelFormat(gl_context->wnd->dc, pixfmt, &pfd))
	{
		SYSCALL_FAILED("SetPixelFormat");
		goto err_exit;
	}
	gl_context->wgl_context = wglCreateContext(gl_context->wnd->dc);
	if(!gl_context->wgl_context)
	{
		SYSCALL_FAILED("wglCreateContext");
		goto err_exit;
	}
	if(!wglMakeCurrent(gl_context->wnd->dc, gl_context->wgl_context))
	{
		SYSCALL_FAILED("wglMakeCurrent");
		goto err_exit;
	}
	gl_context->vendor = strDup((char_p)glGetString(GL_VENDOR));
	gl_context->renderer = strDup((char_p)glGetString(GL_RENDERER));
	gl_context->version = strDup((char_p)glGetString(GL_VERSION));
	gl_context->extensions = strDup((char_p)glGetString(GL_EXTENSIONS));
	archWinPoll(gl_context->wnd, false);
	aspect = (GLfloat)gl_context->wnd->width /
		(GLfloat)gl_context->wnd->height;
	glViewport(0, 0, gl_context->wnd->width, gl_context->wnd->height);
	glMatrixMode(GL_PROJECTION);
	glLoadIdentity();
	glOrtho(-aspect, aspect, -1, 1, 1, 3);
	glMatrixMode(GL_MODELVIEW);
	glLoadIdentity();
	return gl_context;

err_exit:;
	win32_shutdown_gl(gl_context);
	return NULL;
}

vec_t alpha = 0, beta = 0;

int main(void)
{
	ui_event_t ev;
	arch_window_p w;
    win32_gl_context_p gl_context;
	vboolean enable_kb_repeat = false;

	logStartConsoleLog();
	logUpperLevel(LOG_TRACE);


	gl_context = win32_init_gl("test",300,50,640,480,false,false,false,false,false);
//	gl_context = win32_init_gl("test",0,0,1280,1024,true,false,false,false,false);
	assert(gl_context);
/*	win32_shutdown_gl(gl_context);
	memPrintStats();
	return 0;*/
	
	w = gl_context->wnd;

	glClearColor(0,0.5,0.5,0.5);

	for(;;)
	{
		archWinPoll(w, true);

		while(w->msg_que)
		{
			DEQ_POP_BACK(event,w->msg_que,&ev);
			dump_event(0, &ev);

			if(ev.type == ui_window_expose_event || autowarp)
			{
				glClear(GL_COLOR_BUFFER_BIT);
				glPushMatrix();
				glTranslatef(0, 0, -2);

				if(autowarp)
				{
					glRotatef(alpha += ev.client_dx * 0.1,0,1,0);
					glRotatef(beta += ev.client_dy * 0.1,1,0,0);
				}

				glScalef(0.01, 0.01, 1);
				glBegin(GL_TRIANGLES);
					glColor3f(1, 0, 0);
					glVertex3f(0, -50, 0);
					glColor3f(0, 1, 0);
					glVertex3f(-50, 50, 0);
					glColor3f(0, 0, 1);
					glVertex3f(50, 50, 0);
				glEnd();
				
				glPopMatrix();
				win32_end_frame(gl_context);
				if(!autowarp)
					continue;
			}
			if(ev.type != ui_key_event)
				continue;
			switch(ev.key.scan)
			{
				case sc_escape:
					if(!ev.key.press)
						goto close_win;
					break;
				case sc_s:
					archWinShow(w, false);
					archWinPoll(w, false);
					while(w->msg_que)
					{
						DEQ_POP_BACK(event,w->msg_que,&ev);
						dump_event(0, &ev);
					}
					timeDelay(3000);
					archWinShow(w, true);
					break;
				case sc_p:
					archWinShowPointer(w, true);
					break;
				case sc_o:
					archWinShowPointer(w, false);
					break;
				case sc_a:
					if(ev.key.press)
						archWinAutoRepeat(w,enable_kb_repeat = !enable_kb_repeat);
					break;
				case sc_w:
					if(ev.key.press)
						archWinAutoWarp(w,autowarp = !autowarp);
					break;
				default:
					break;
			}
		}
	}

close_win:;
	win32_shutdown_gl(gl_context);

	memPrintStats();
	return 0;
}
